from types import TracebackType

class PWMError(IOError): ...

class PWM:
    def __init__(self, chip: int, channel: int) -> None: ...
    def __del__(self) -> None: ...
    def __enter__(self) -> PWM: ...  # noqa: Y034
    def __exit__(self, t: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None) -> None: ...
    def close(self) -> None: ...
    enabled: bool
    def enable(self) -> None: ...
    def disable(self) -> None: ...
    @property
    def devpath(self) -> str: ...
    @property
    def chip(self) -> int: ...
    @property
    def channel(self) -> int: ...
    period_ns: int
    duty_cycle_ns: int
    period: int | float
    duty_cycle: int | float
    frequency: int | float
    polarity: str
